import unittest
from serum import immutable


class C:
    a = immutable('a')


class ImmutableTest(unittest.TestCase):
    def test_immutable_fields_are_immutable(self):
        c = C()
        with self.assertRaises(AttributeError):
            c.a = 'b'

    def test_immutable_fields_return_correct_value(self):
        c = C()
        self.assertEqual(c.a, 'a')



from serum._environment import provide


class LazyDependency:
    def __init__(self, dependency):
        self.__dependency = dependency

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return provide(self.__dependency, instance)

    def __set__(self, instance, value):
        raise AttributeError('Can\'t set injected attribute')



from typing import TypeVar
from abc import ABCMeta
import inspect
from .exceptions import InvalidComponent


def _more_than_one_parameter_to_init(dct):
    if '__init__' not in dct:
        return False
    signature = inspect.signature(dct['__init__'])
    return len(signature.parameters) > 1


class _ComponentMeta(ABCMeta):
    def __new__(mcs, name, bases, dct):
        if _more_than_one_parameter_to_init(dct):
            raise InvalidComponent(
                "__init__ method in Components can only take 1 parameter"
            )
        return ABCMeta.__new__(mcs, name, bases, dct)


class Component(metaclass=_ComponentMeta):
    """
    Base class for all injectable types.
    Prevents __init__ method with more than one parameter:

    class MyComponent(Component):  # raises: InvalidComponent
        def __init__(self, a):
            pass

    In addition, Components can be abstract, meaning they
    work with the built abc module:

    class MyAbstractComponent(Component):
        @abstractmethod
        def method(self):
            pass

    MyAbstractComponent()  # raises: TypeError
    """


class Singleton(Component):
    """
    Base class for singleton types. References to singletons always refer to
    the same instance:

    class ExpensiveObject(Singleton):
        pass

    class Dependent:
        expensive_instance = inject(ExpensiveObject)

    with Environment():
        instance1 = Dependent()
        instance2 = Dependent()
        assert instance1.expensive_instance == instance2.expensive_instance
    """
    pass


__all__ = ['Component', 'Singleton']



import unittest
import os

from serum import Environment, match
from serum.exceptions import UnknownEnvironment


class MatchTests(unittest.TestCase):
    def tearDown(self):
        if 'TEST_ENV' in os.environ:
            del os.environ['TEST_ENV']

    def test_match_returns_correct_env(self):
        env1 = Environment()
        env2 = Environment()
        os.environ['TEST_ENV'] = 'ENV1'
        env = match(environment_variable='TEST_ENV', ENV1=env1, ENV2=env2)
        self.assertIs(env, env1)
        os.environ['TEST_ENV'] = 'ENV2'
        env = match(environment_variable='TEST_ENV', ENV1=env1, ENV2=env2)
        self.assertIs(env, env2)

    def test_match_gets_default(self):
        default = Environment()
        env1 = Environment()
        env = match(environment_variable='TEST_ENV', default=default, ENV1=env1)
        self.assertEqual(env, default)

    def test_match_fails_when_no_default_and_no_env(self):
        env1 = Environment()
        with self.assertRaises(UnknownEnvironment):
            match(environment_variable='TEST_ENV', env1=env1)

    def test_match_fails_with_unknown_environment(self):
        os.environ['TEST_ENV'] = 'unknown'
        with self.assertRaises(UnknownEnvironment):
            match(environment_variable='TEST_ENV')



import unittest
from serum import mock, Component, inject, Environment
from serum.exceptions import NoEnvironment


class SomeComponent(Component):
    def method(self):
        return 'some value'


class SomeCallableComponent(Component):
    def __call__(self):
        return 'some value'


class Depenedent:
    some_component = inject(SomeComponent)
    some_callable_component = inject(SomeCallableComponent)


class MockTests(unittest.TestCase):
    def test_mock_always_replaces_component(self):
        with Environment():
            some_component_mock = mock(SomeComponent)
            some_component_mock.method.return_value = 'some other value'
            d = Depenedent()
            self.assertIs(d.some_component, some_component_mock)
            self.assertEqual(d.some_component.method(), 'some other value')

    def test_mocks_are_reset_after_context_exit(self):
        with Environment():
            some_component_mock = mock(SomeComponent)
            d = Depenedent()
            self.assertIs(some_component_mock, d.some_component)

        with Environment():
            d = Depenedent()
            self.assertIsNot(some_component_mock, d.some_component)
            self.assertIsInstance(d.some_component, SomeComponent)

    def test_cant_register_mocks_outside_environment(self):
        with self.assertRaises(NoEnvironment):
            mock(SomeComponent)

    def test_mock_is_specced(self):
        with Environment():
            some_component_mock = mock(SomeComponent)
            self.assertIsInstance(some_component_mock, SomeComponent)
            with self.assertRaises(AttributeError):
                some_component_mock.bad_method()
            with self.assertRaises(TypeError):
                some_component_mock()
            some_callable_component = mock(SomeCallableComponent)
            some_callable_component.return_value = 'mocked value'
            self.assertEqual(some_callable_component(), 'mocked value')

    def test_mock_replaces_named_value(self):
        class Dependency:
            def method(self):
                pass

        with Environment(key=Dependency()):
            mock_dependency = mock('key')
            mock_dependency.method.return_value = 'value'
            with self.assertRaises(AttributeError):
                mock_dependency.no_such_method()
            injected = inject('key')
            self.assertEqual(injected, mock_dependency)
            self.assertEqual(mock_dependency.method(), 'value')



from abc import abstractmethod, abstractstaticmethod, abstractclassmethod

from ._component import *
from ._environment import *
from ._functions import *


def load_ipython_extension(_):
    try:
        from ipython_environment import environment
        environment.__enter__()
    except ModuleNotFoundError:
        pass


def unload_ipython_extension(_):
    try:
        from ipython_environment import environment
        environment.__exit__(None, None, None)
    except ModuleNotFoundError:
        pass



import unittest
from serum import Component, abstractmethod
from serum.exceptions import InvalidComponent


class ComponentTests(unittest.TestCase):
    def test_component_init_only_one_parameter(self):
        class SomeComponent(Component):
            def __init__(self):
                pass

        with self.assertRaises(InvalidComponent):
            class SomeComponent(Component):
                def __init__(self, a):
                    pass

    def test_component_can_be_abstract(self):
        class AbstractComponent(Component):
            @abstractmethod
            def test(self):
                pass

        with self.assertRaises(TypeError):
            AbstractComponent()



from copy import copy, deepcopy
from unittest.mock import create_autospec, MagicMock
from functools import wraps
from typing import Type, TypeVar, Set, Union, Dict
from weakref import WeakKeyDictionary

from .exceptions import (
    InvalidDependency,
    NoEnvironment,
    UnregisteredDependency,
    AmbiguousDependencies,
    CircularDependency)
from ._component import Component, Singleton
import threading
import inspect
from collections import Counter

C = TypeVar('C', bound=Component)


class _LocalStorage(threading.local):
    def __init__(self):
        self.current_env: Environment = None


class _EnvironmentState(threading.local):
    def __init__(self):
        self.pending: Set[Type[C]] = set()
        self.old_current: Environment = None
        self.mocks: Dict[Type[C], MagicMock] = dict()
        self.singletons: Dict[Type[C], C] = dict()
        self.instances: Dict[object, Dict[Type[C], C]] = WeakKeyDictionary()

    def __deepcopy__(self, memodict):
        new = _EnvironmentState()
        new.pending = copy(self.pending)
        new.old_current = copy(self.old_current)
        new.mocks = copy(self.mocks)
        new.singletons = copy(self.mocks)
        new.instances = copy(self.instances)
        return new


class Environment:
    """
    Context manager/decorator for providing instances of Component:

    with Environment(MyDependency):
        NeedsDependency()
    or

    @Environment(MyDependency):
    def fun():
        NeedsDependency()
    """
    __local_storage = _LocalStorage()

    @staticmethod
    def current_env() -> 'Environment':
        env = Environment.__local_storage.current_env
        if env is None:
            raise NoEnvironment(
                'Can\'t provide dependencies outside an environment'
            )
        return env

    @staticmethod
    def mock(dependency: Union[str, Type[C]]):
        try:
            current_env = Environment.current_env()
        except NoEnvironment:
            raise NoEnvironment(
                'Can\'t register mock outside environment'
            )
        if isinstance(dependency, str):
            value = current_env[dependency]
            mock = create_autospec(value)
            current_env.__named_dependencies[dependency] = mock
            return mock
        mock = create_autospec(dependency, instance=True)
        current_env.__state.mocks[dependency] = mock
        return mock

    @staticmethod
    def _set_current_env(env: 'Environment'):
        Environment.__local_storage.current_env = env

    def __init__(self, *args: Type[C], **kwargs) -> None:
        """
        Construct a new environment
        :param args: Components to provide in this environment
        :param kwargs: Named dependencies to provide in this environment
        """
        self.__registry: Set[Type[C]] = set()
        self.__state: _EnvironmentState = _EnvironmentState()
        self.__named_dependencies = kwargs
        self.__old_current = None
        for c in args:
            self.__use(c)

    def __getitem__(self, item: str):
        """
        Get named dependency in this environment
        :param item: Name of dependency
        :return: dependency
        """
        try:
            return self.__named_dependencies[item]
        except KeyError:
            raise KeyError(f'No named dependency: {item}')

    @property
    def pending(self) -> Set[Type[C]]:
        return self.__state.pending

    def get_mock(self, component: Type[C]) -> MagicMock:
        return self.__state.mocks[component]

    def is_mocked(self, component: Type[C]) -> bool:
        return component in self.__state.mocks

    def __use(self, component: Type[C]) -> 'Environment':
        if not issubclass(component, Component):
            raise InvalidDependency(
                'Attempt to register type that is not a Component: {}'.format(
                    str(component)
                )
            )
        self.__registry.add(component)
        return self

    def __contains__(self, component: Type[C]) -> bool:
        """
        Test if a Component is registered in this environment
        :param component: Component to test
        :return: True if component is registered in this environment else False
        """
        return component in self.__registry

    def __call__(self, f):
        """
        Decorate a function to run in this environment
        :param f: function to decorate
        :return: decorated function
        """
        @wraps(f)
        def run_in(*args, **kwargs):
            with self:
                f(*args, **kwargs)
        return run_in

    def __iter__(self):
        """
        Iterate over the components registered in this environment
        :return: Iterator of components
        """
        return iter(self.__registry)

    def __or__(self, other: 'Environment') -> 'Environment':
        """
        Combine this environment with another, such that the new environment
        can provide all components in both environments
        :param other: Environment to combine with this one
        :return: New environment with components from this and the other
                 environment
        """
        new_registry = self.__registry | other.__registry
        return Environment(*new_registry)

    def __enter__(self):
        """
        Register this environment as the current environment in this thread
        :return:
        """
        try:
            self.__old_current = Environment.current_env()
        except NoEnvironment:
            pass
        if self.__old_current is not None:
            old_state = self.__old_current.__copy_state()
            self.__set_state(old_state)
        Environment._set_current_env(self)
        return self

    def __set_state(self, state):
        self.__state = state

    def __copy_state(self):
        return deepcopy(self.__state)

    def has_singleton_instance(self, singleton_type):
        return singleton_type in self.__state.singletons

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        De-register this environment as the current environment in this thread
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.__state.__init__()
        Environment._set_current_env(self.__old_current)
        self.__old_current = None

    @staticmethod
    def provide(component: Type[C], caller: object) -> Union[C, MagicMock]:
        """
        Provide a component in this environment
        :param caller:
        :param component: The type to provide
        :return: Instance of the most specific subtype of component
                 in this environment
        """
        current_env = Environment.current_env()

        def singleton(singleton_type: Type[C]) -> C:
            if current_env.has_singleton_instance(singleton_type):
                return current_env.get_singleton(singleton_type)
            else:
                singleton_instance = singleton_type()
                current_env.add_singleton(
                    singleton_type,
                    singleton_instance
                )
                return singleton_instance

        def instance(component_type: Type[C]) -> C:
            component_instance = component_type()
            current_env.set_instance(component, caller, component_instance)
            return component_instance

        def is_singleton(st):
            return issubclass(st, Singleton)

        def instantiate(component_type: Type[C]) -> C:
            if is_singleton(component_type):
                return singleton(component_type)
            return instance(component_type)

        if current_env.is_mocked(component):
            return current_env.get_mock(component)
        if current_env.has_instance(component, caller):
            return current_env.get_instance(component, caller)
        try:
            subtype = Environment._find_subtype(component)
            if subtype in current_env.pending:
                raise CircularDependency(
                    'Circular dependency encountered while injecting {} in {}'.format(
                        str(component),
                        str(caller)
                    )
                )
            current_env.pending.add(subtype)
            subtype_instance = instantiate(subtype)
            current_env.pending.remove(subtype)
            return subtype_instance
        except UnregisteredDependency:
            try:
                return instantiate(component)
            except TypeError:
                raise UnregisteredDependency(
                    'No concrete implementation of {} found'.format(
                        str(component)
                    )
                )

    @staticmethod
    def _find_subtype(component: Type[C]) -> Type[C]:
        def mro_distance(subtype: Type[C]) -> int:
            mro = inspect.getmro(subtype)
            return mro.index(component)

        subtypes = [c for c in Environment.current_env()
                    if issubclass(c, component)]
        distances = [mro_distance(subtype) for subtype in subtypes]
        counter = Counter(distances)
        if any(count > 1 for count in counter.values()):
            ambiguous = [str(subtype) for subtype in subtypes
                         if counter[mro_distance(subtype)] > 1]
            message = ('Attempt to inject type {} with '
                       'equally specific provided subtypes: {}')
            message = message.format(
                str(component),
                ', '.join(ambiguous)
            )
            raise AmbiguousDependencies(message)
        if not subtypes:
            raise UnregisteredDependency(
                'Unregistered dependency: {}'.format(str(component))
            )
        return max(subtypes, key=mro_distance)

    def get_singleton(self, singleton_type):
        return self.__state.singletons[singleton_type]

    def add_singleton(self, singleton_type, instance):
        self.__state.singletons[singleton_type] = instance

    def has_instance(self, component, caller):
        return caller in self.__state.instances and component in self.__state.instances[caller]

    @property
    def instances(self):
        return self.__state.instances

    def get_instance(self, component, caller):
        return self.__state.instances[caller][component]

    def set_instance(self, component, caller, component_instance):
        if caller not in self.__state.instances:
            self.__state.instances[caller] = {}
        self.__state.instances[caller][component] = component_instance


__all__ = ['Environment']


def provide(dependency, caller):
    if isinstance(dependency, str):
        return Environment.current_env()[dependency]
    return Environment.provide(dependency, caller)


class NoEnvironment(Exception):
    """
    Exception type to signal attempts to call functions that require
    an environment outside an environment
    """
    pass


class InvalidDependency(Exception):
    """
    Exception type to signal attempts to inject/register types that do
    not inherit from Component
    """
    pass


class UnregisteredDependency(Exception):
    """
    Exception type to signal failure to find an appropriate concrete type
    in an environment
    """
    pass


class InvalidComponent(Exception):
    """
    Exception type to signal bad declaration of a Component subtype
    """
    pass


class AmbiguousDependencies(Exception):
    """
    Exception type to signal registering conflicting types in an environment
    """
    pass


class CircularDependency(Exception):
    """
    Exception type to signal attempt to inject dependency with circular
    dependencies
    """
    pass


class UnknownEnvironment(Exception):
    pass


import unittest
import sys
from serum import Environment, load_ipython_extension, unload_ipython_extension
from serum.exceptions import NoEnvironment


class MockIpythonEnvironment:
    environment = Environment()


class IPythonExtensionTests(unittest.TestCase):

    def test_load_extension_no_ipython_environment(self):
        self.assertEqual(load_ipython_extension(None), None)

    def test_load_extension(self):
        sys.modules['ipython_environment'] = MockIpythonEnvironment
        load_ipython_extension(None)
        self.assertEqual(
            MockIpythonEnvironment.environment,
            Environment.current_env()
        )
        unload_ipython_extension(None)
        with self.assertRaises(NoEnvironment):
            Environment.current_env()
        sys.modules['ipython_environment'] = None

    def test_unload_extension_no_ipython_environment(self):
        self.assertEqual(unload_ipython_extension(None), None)


from setuptools import setup
from scripts.release import read_version

version = read_version('version.txt')
url = 'https://github.com/suned/serum/archive/{}.tar.gz'.format('v' + version)
setup(
  name='serum',
  packages=['serum'],
  version=version,
  description='Dependency Injection library for Python 3',
  author='Sune Debel',
  license='MIT',
  author_email='sd@dybro-debel.dk',
  python_requires='>=3.5',
  url='https://github.com/suned/serum',
  download_url=url,
  keywords=['dependency-injection', 'solid', 'inversion-of-control'],
  classifiers=[],
)



import unittest

import gc

from queue import Queue

from serum import Environment, Component, abstractmethod, Singleton, inject
from serum.exceptions import InvalidDependency, UnregisteredDependency, \
    NoEnvironment, AmbiguousDependencies, CircularDependency
import threading


class SomeComponent(Component):
    pass


class SomeOtherComponent(Component):
    pass


class AbstractComponent(Component):
    @abstractmethod
    def m(self):
        pass


class ConcreteComponent(AbstractComponent):
    def m(self):
        pass


class AlternativeComponent(AbstractComponent):
    def m(self):
        pass


class SomeSingleton(Singleton):
    pass


class Key:
    pass


class Dependent:
    some_singleton = inject(SomeSingleton)
    some_component = inject(SomeComponent)


def check_garbage(self, e):
    def _(phase, info):
        if phase == 'stop':
            self.assertEqual(e.instances, {})
    gc.callbacks.append(_)
    gc.collect()
    gc.callbacks.remove(_)


class EnvironmentTests(unittest.TestCase):

    def test_cant_register_non_component(self):
        class NotAComponent:
            pass
        with self.assertRaises(InvalidDependency):
            Environment(NotAComponent, self)

    def test_can_register_component(self):
        e = Environment(SomeComponent)
        self.assertTrue(SomeComponent in e)

    def test_environment_provides_concrete_component(self):
        with Environment():
            c = Environment.provide(SomeComponent, self)
            self.assertIsInstance(c, SomeComponent)

    def test_environment_cant_provide_abstract_component(self):
        with Environment():
            with self.assertRaises(UnregisteredDependency):
                Environment.provide(AbstractComponent, self)

    def test_environment_provides_concrete_subclass(self):
        with Environment(ConcreteComponent):
            c = Environment.provide(AbstractComponent, self)
            self.assertIsInstance(c, AbstractComponent)
            self.assertIsInstance(c, ConcreteComponent)

    def test_environment_provides_correct_implementation(self):
        with Environment(ConcreteComponent):
            c = Environment.provide(AbstractComponent, self)
            self.assertIsInstance(c, AbstractComponent)
            self.assertIsInstance(c, ConcreteComponent)
        with Environment(AlternativeComponent):
            c = Environment.provide(AbstractComponent, self)
            self.assertIsInstance(c, AbstractComponent)
            self.assertIsInstance(c, AlternativeComponent)

    def test_intersection(self):
        e1 = Environment(SomeComponent)
        e2 = Environment(ConcreteComponent)
        e3 = e1 | e2
        self.assertIn(SomeComponent, e3)
        self.assertIn(ConcreteComponent, e3)

    def test_decorater(self):
        test_environment = Environment(SomeComponent)

        @test_environment
        def test():
            component = Environment.provide(SomeComponent, self)
            self.assertIsInstance(component, SomeComponent)

        test()

    def test_new_environment_in_thread(self):
        def test():
            with Environment(AlternativeComponent):
                c1 = Environment.provide(AbstractComponent, self)
                self.assertIsInstance(c1, AlternativeComponent)

        with Environment(ConcreteComponent):
            threading.Thread(target=test).start()
            c2 = Environment.provide(AbstractComponent, self)
            self.assertIsInstance(c2, ConcreteComponent)

    def test_same_environment_in_thread(self):
        def test():
            with self.assertRaises(NoEnvironment):
                Environment.provide(AbstractComponent, self)

        with Environment(ConcreteComponent):
            threading.Thread(target=test).start()

    def test_nested_environments(self):
        key = Key()
        with Environment(ConcreteComponent):
            c1 = Environment.provide(AbstractComponent, self)
            self.assertIsInstance(c1, ConcreteComponent)
            with Environment(AlternativeComponent):
                c2 = Environment.provide(AbstractComponent, self)
                c3 = Environment.provide(AbstractComponent, key)
                self.assertIs(c1, c2)
                self.assertIsInstance(c3, AlternativeComponent)
            c4 = Environment.provide(AbstractComponent, key)
            c5 = Environment.provide(AbstractComponent, self)
            self.assertIsNot(c3, c4)
            self.assertIs(c1, c5)

    def test_context_manager(self):
        e = Environment()
        with e:
            self.assertIs(Environment.current_env(), e)
        with self.assertRaises(NoEnvironment):
            Environment.current_env()

    def test_missing_named_dependency(self):
        e = Environment()
        with self.assertRaises(KeyError):
            _ = e['key']

    def test_getitem(self):
        e = Environment(key='value')
        self.assertEqual(e['key'], 'value')

    def test_environment_gets_most_specific(self):
        class ConcreteComponentSub(ConcreteComponent):
            pass

        with Environment(ConcreteComponent, ConcreteComponentSub):
            c = Environment.provide(AbstractComponent, self)
            self.assertIsInstance(c, ConcreteComponentSub)

    def test_fails_with_ambiguous_dependencies(self):
        with Environment(ConcreteComponent, AlternativeComponent):
            with self.assertRaises(AmbiguousDependencies):
                Environment.provide(AbstractComponent, self)

    def test_singleton_is_always_same_instance(self):
        with Environment():
            s1 = Environment.provide(SomeSingleton, self)
            s2 = Environment.provide(SomeSingleton, object())
            self.assertIs(s1, s2)

    def test_instances_are_cached(self):
        with Environment():
            s1 = Environment.provide(SomeComponent, self)
            s2 = Environment.provide(SomeComponent, self)
            self.assertIs(s1, s2)

    def test_more_than_one_instance_in_cache(self):
        with Environment():
            s1 = Environment.provide(SomeComponent, self)
            s2 = Environment.provide(SomeOtherComponent, self)
            s3 = Environment.provide(SomeComponent, self)
            s4 = Environment.provide(SomeOtherComponent, self)
            self.assertIs(s1, s3)
            self.assertIs(s2, s4)

    def test_circular_dependency(self):
        class AbstractA(Component):
            pass

        class AbstractB(Component):
            pass

        class A(AbstractA):
            b = inject(AbstractB)

            def __init__(self):
                self.b

        class B(AbstractB):
            a = inject(AbstractA)

            def __init__(self):
                self.a

        class Dependent:
            a = inject(AbstractA)

        with Environment(A, B):
            with self.assertRaises(CircularDependency):
                Dependent().a

    def test_subtype_is_singleton(self):
        class SomeComponentSingleton(SomeComponent, Singleton):
            pass
        with Environment(SomeComponentSingleton):
            s1 = Environment.provide(SomeComponent, object())
            s2 = Environment.provide(SomeComponent, object())
            self.assertIs(s1, s2)
            s3 = Environment.provide(SomeComponentSingleton, object())
            self.assertIs(s1, s3)

    def test_garbage_collection(self):
        with Environment() as e:
            d = Dependent()
            _ = d.some_component
            gc.collect()
            self.assertTrue(e.has_instance(SomeComponent, d))
            del d
            check_garbage(self, e)

    def test_scope(self):
        environment = Environment()
        d = Dependent()

        with environment:
            sc1 = d.some_component
            ss1 = d.some_singleton
        with environment:
            sc2 = d.some_component
            ss2 = d.some_singleton
        self.assertIsNot(sc1, sc2)
        self.assertIsNot(ss1, ss2)

    def test_scope_multi_threaded(self):
        environment = Environment()
        d = Dependent()
        q = Queue()
        c = threading.Condition()

        with environment:
            some_component1 = d.some_component
            singleton1 = d.some_singleton

        def t1():
            with c:
                c.wait()
            some_component2 = q.get()
            singleton2 = q.get()
            with environment:
                q.put(d.some_component)
                q.put(d.some_singleton)
                with c:
                    c.notify()
                self.assertIsNot(d.some_component, some_component2)
                self.assertIsNot(d.some_singleton, singleton2)
                self.assertIsNot(d.some_component, some_component1)
                self.assertIsNot(d.some_singleton, singleton1)

        def t2():
            with environment:
                q.put(d.some_component)
                q.put(d.some_singleton)
                with c:
                    c.notify()
                    c.wait()
                some_component3 = q.get()
                singleton3 = q.get()
                self.assertIsNot(d.some_component, some_component3)
                self.assertIsNot(d.some_singleton, singleton3)
                self.assertIsNot(d.some_component, some_component1)
                self.assertIsNot(d.some_singleton, singleton1)

        threading.Thread(target=t1).start()
        threading.Thread(target=t2).start()



from typing import TypeVar, Type, cast, overload, Any, Union
from unittest.mock import MagicMock

import os

from serum._lazy_dependency import LazyDependency
from ._environment import provide
from .exceptions import InvalidDependency, UnknownEnvironment, NoEnvironment
from ._component import Component
from ._environment import Environment

C = TypeVar('C', bound=Component)
T = TypeVar('T')


class Key:
    pass


@overload
def inject(dependency: Type[C]) -> C:
    pass  # pragma: no cover


@overload
def inject(dependency: str) -> Any:
    pass  # pragma: no cover


def inject(dependency):
    """
    Inject a dependency
    :param dependency: The Component type or key to inject
    :return: Instantiated Component or key
    """
    if not (isinstance(dependency, str) or issubclass(dependency, Component)):
        raise InvalidDependency(
            'Attempt to inject dependency that is not a Component or str: {}'.format(
                str(dependency)
            )
        )

    try:
        return provide(dependency, Key())
    except NoEnvironment:
        return LazyDependency(dependency)


def immutable(value: T) -> T:
    """
    Add a static constant to a class.
    :param value: The value to add
    :return: property that returns value
    """
    return cast(T, property(fget=lambda _: value))


def mock(dependency: Union[str, Type[C]]) -> MagicMock:
    """
    Mock a component in the current environment
    :param dependency: The type to mock
    :return: unittest.mock.MagicMock instance that replaces component in this
             environment
    """
    return Environment.mock(dependency)


def match(environment_variable: str,
          default: Environment = None,
          **environments) -> Environment:
    """
    Match environment variable with Environment
    :param environment_variable: environment variable to match environment against
    :param default: default environment when no value is
                    assigned to environment_variable
    :param environments: kwargs of environments to use for different values of
                         environment_variable
    :return: Environment matched against the value of environment_variable
             or default Environment is no value is assigned
    """
    environment = os.environ.get(environment_variable, 'default')
    if environment == 'default':
        if default is None:
            raise UnknownEnvironment(
                'No environment specified and no default environment'
            )
        return default
    try:
        return environments[environment]
    except KeyError:
        raise UnknownEnvironment(environment)


__all__ = ['inject', 'immutable', 'mock', 'match']



import unittest
from serum import inject, Component, Environment, abstractmethod
from serum._lazy_dependency import LazyDependency
from serum.exceptions import NoEnvironment, UnregisteredDependency, \
    InvalidDependency


class SomeComponent(Component):
    pass


class AbstractComponent(Component):
    @abstractmethod
    def abstract(self):
        pass


class ConcreteComponent(AbstractComponent):
    def abstract(self):
        pass


class AlternativeComponent(AbstractComponent):
    def abstract(self):
        pass


class Chain(Component):
    some_component = inject(SomeComponent)


class Dependent:
    some_component = inject(SomeComponent)
    abstract_component = inject(AbstractComponent)
    chain = inject(Chain)
    value = inject('key')


class InjectTests(unittest.TestCase):
    def test_inject_fails_outside_environment(self):
        d = Dependent()
        with self.assertRaises(NoEnvironment):
            _ = d.some_component

    def test_inject_gets_concrete_component(self):
        with Environment():
            d = Dependent()
            self.assertIsInstance(d.some_component, SomeComponent)

    def test_injected_component_is_immutable(self):
        d = Dependent()
        with Environment():
            with self.assertRaises(AttributeError):
                d.some_component = 'test'

    def test_inside_environment(self):
        with Environment():
            self.assertIsInstance(inject(SomeComponent), SomeComponent)

    def test_inject_string(self):
        with Environment(key='value'):
            self.assertEqual(inject('key'), 'value')

    def test_static_dependency(self):
        self.assertIsInstance(Dependent.some_component, LazyDependency)

    def test_inject_cant_get_abstract_component(self):
        d = Dependent()
        with Environment():
            with self.assertRaises(UnregisteredDependency):
                _ = d.abstract_component

    def test_inject_can_get_concrete_component(self):
        d = Dependent()
        with Environment(ConcreteComponent):
            self.assertIsInstance(d.abstract_component, AbstractComponent)
            self.assertIsInstance(d.abstract_component, ConcreteComponent)

    def test_inject_provides_correct_implementation(self):
        d = Dependent()
        with Environment(ConcreteComponent):
            self.assertIsInstance(d.abstract_component, AbstractComponent)
            self.assertIsInstance(d.abstract_component, ConcreteComponent)
        with Environment(AlternativeComponent):
            self.assertIsInstance(d.abstract_component, AbstractComponent)
            self.assertIsInstance(d.abstract_component, AlternativeComponent)

    def test_injection_chaining(self):
        d = Dependent()
        with Environment():
            self.assertIsInstance(d.chain, Chain)
            self.assertIsInstance(d.chain.some_component, SomeComponent)

    def test_inject_non_component_fails(self):
        class Test:
            pass
        with self.assertRaises(InvalidDependency):
            inject(Test)

    def test_injected_is_always_same_instance(self):
        with Environment():
            d1 = Dependent()
            self.assertIs(d1.some_component, d1.some_component)

    def test_injected_are_different_instances(self):
        with Environment():
            d1 = Dependent()
            d2 = Dependent()
            self.assertIsNot(d1.some_component, d2.some_component)

    def test_lazy_named_dependency(self):
        with Environment(key='value'):
            self.assertEqual(Dependent().value, 'value')
